import plotly.express as px

dark_blue = "rgb(40, 60, 70)"
light_blue = "rgb(200, 230, 250)"
plot_background_blue = "rgb(240, 250, 255)"
transparent = "rgba(255, 255, 255, 0)"


def get_rubicon_colorscale(num_colors, low=0.33):
    if num_colors < 2:
        num_colors = 2

    return px.colors.sample_colorscale(
        "Blues",
        num_colors,
        low=low,
    )



import dash_bootstrap_components as dbc
from dash import html

from rubicon_ml.viz.base import VizBase

COL_WIDTH_LOOKUP = {1: 12, 2: 6, 3: 4, 4: 3}


class Dashboard(VizBase):
    """Compose visualizations into a dashboard to view multiple widgets at once.

    Parameters
    ----------
    experiments : list of rubicon_ml.client.experiment.Experiment
        The experiments to visualize.
    widgets : list of lists of superclasses of rubicon_ml.viz.base.VizBase, optional
        The widgets to compose in this dashboard. The widgets should be instantiated
        without experiments prior to passing as an argument to `Dashboard`. Defaults
        to a stacked layout of an ExperimentsTable and a MetricCorrelationPlot.
    link_experiment_table : bool, optional
        True to enable the callbacks that allow instances of `ExperimentsTable` to
        update the experiment inputs of the other widgets in this dashboard. False
        otherwise. Defaults to True.
    """

    def __init__(self, experiments, widgets=None, link_experiment_table=True):
        super().__init__(dash_title="dashboard")

        self.experiments = experiments
        self.link_experiment_table = link_experiment_table

        if widgets is None:
            from rubicon_ml.viz import ExperimentsTable, MetricCorrelationPlot

            self.widgets = [
                [ExperimentsTable(is_selectable=True)],
                [MetricCorrelationPlot()],
            ]
        else:
            self.widgets = widgets

    @property
    def layout(self):
        """Defines the layout for the dashboard.

        Compiles the figures in `self.widgets` based on their positions
        in the given input list.
        """
        dashboard_rows = []
        for row in self.widgets:
            width = COL_WIDTH_LOOKUP[len(row)]

            row_widgets = []
            for widget in row:
                row_widgets.append(dbc.Col(widget.layout, width=width))

            dashboard_rows.append(dbc.Row(row_widgets))

        dashboard_container = html.Div(dashboard_rows)

        return dashboard_container

    def load_experiment_data(self):
        """Load the experiment data required for the dashboard.

        Loads the experiment data for each widget in `self.widgets`.
        """
        for row in self.widgets:
            for widget in row:
                widget.experiments = self.experiments
                widget.load_experiment_data()

    def register_callbacks(self):
        for row in self.widgets:
            for widget in row:
                widget.app = self.app
                widget.register_callbacks(self.link_experiment_table)



import functools
import logging
import traceback
import warnings

FAILURE_MODE = "raise"
FAILURE_MODES = ["log", "raise", "warn"]
TRACEBACK_CHAIN = False
TRACEBACK_LIMIT = None


def set_failure_mode(failure_mode, traceback_chain=False, traceback_limit=None):
    """Set the failure mode.

    Parameters
    ----------
    failure_mode : str
        The name of the failure mode to set. "raise" to raise all exceptions,
        "log" to catch all exceptions and log them via `logging.error`, "warn"
        to catch all exceptions and re-raise them as warnings via `warnings.warn`.
        Defaults to "raise".
    traceback_chain : bool, optional
        True to display each error in the traceback chain when logging or warning,
        False to display only the first. Defaults to False.
    traceback_limit : int, optional
        The depth of the traceback displayed when logging or warning. 0 to display
        only the error's text, each increment shows another line of the traceback.
    """
    global FAILURE_MODE
    global TRACEBACK_CHAIN
    global TRACEBACK_LIMIT

    if failure_mode not in FAILURE_MODES:
        raise ValueError(f"`failure_mode` must be one of {FAILURE_MODES}")

    FAILURE_MODE = failure_mode
    TRACEBACK_CHAIN = traceback_chain
    TRACEBACK_LIMIT = traceback_limit


def failsafe(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if FAILURE_MODE == "raise":
                raise e
            elif FAILURE_MODE == "warn":
                warnings.warn(traceback.format_exc(limit=TRACEBACK_LIMIT, chain=TRACEBACK_CHAIN))
            elif FAILURE_MODE == "log":
                logging.error(traceback.format_exc(limit=TRACEBACK_LIMIT, chain=TRACEBACK_CHAIN))

    return wrapper



import os
import random

import numpy as np
import pandas as pd
import pytest

from rubicon_ml.repository import MemoryRepository


class MockCompletedProcess:
    """Use to mock a CompletedProcess result from
    `subprocess.run()`.
    """

    def __init__(self, stdout="", returncode=0):
        self.stdout = stdout
        self.returncode = returncode


@pytest.fixture
def mock_completed_process_empty():
    return MockCompletedProcess(stdout=b"\n")


@pytest.fixture
def mock_completed_process_git():
    return MockCompletedProcess(stdout=b"origin github.com (fetch)\n")


@pytest.fixture
def rubicon_client():
    """Setup an instance of rubicon configured to log to memory
    and clean it up afterwards.
    """
    from rubicon_ml import Rubicon

    rubicon = Rubicon(persistence="memory", root_dir="./")

    # teardown after yield
    yield rubicon
    rubicon.repository.filesystem.rm(rubicon.config.root_dir, recursive=True)


@pytest.fixture
def rubicon_local_filesystem_client():
    """Setup an instance of rubicon configured to log to the
    filesystem and clean it up afterwards.
    """
    from rubicon_ml import Rubicon

    rubicon = Rubicon(
        persistence="filesystem",
        root_dir=os.path.join(os.path.dirname(os.path.realpath(__file__)), "rubicon"),
    )

    # teardown after yield
    yield rubicon
    rubicon.repository.filesystem.rm(rubicon.config.root_dir, recursive=True)


@pytest.fixture
def rubicon_local_filesystem_client_with_project(rubicon_local_filesystem_client):
    rubicon = rubicon_local_filesystem_client

    project_name = "Test Project"
    project = rubicon.get_or_create_project(project_name, description="testing")

    return rubicon, project


@pytest.fixture
def project_client(rubicon_client):
    """Setup an instance of rubicon configured to log to memory
    with a default project and clean it up afterwards.
    """
    rubicon = rubicon_client

    project_name = "Test Project"
    project = rubicon.get_or_create_project(
        project_name, description="In memory project for testing."
    )

    return project


@pytest.fixture
def rubicon_and_project_client(rubicon_client):
    """Setup an instance of rubicon configured to log to memory
    with a default project and clean it up afterwards. Expose
    both the rubicon instance and the project.
    """
    rubicon = rubicon_client

    project_name = "Test Project"
    project = rubicon.get_or_create_project(
        project_name,
        description="In memory project for testing.",
        github_url="test.github.url.git",
    )

    return (rubicon, project)


@pytest.fixture
def rubicon_and_project_client_with_experiments(rubicon_and_project_client):
    """Setup an instance of rubicon configured to log to memory
    with a default project with experiments and clean it up afterwards.
    Expose both the rubicon instance and the project.
    """
    rubicon, project = rubicon_and_project_client

    for e in range(0, 10):
        experiment = project.log_experiment(
            tags=["testing"],
            commit_hash=str(int(e / 3)),
            training_metadata=("training", "metadata"),
        )
        experiment.log_parameter("n_estimators", e + 1)
        experiment.log_feature("year")
        experiment.log_metric("accuracy", (80 + e))

    return (rubicon, project)


@pytest.fixture
def test_dataframe():
    """Create a test dataframe which can be logged to a project or experiment."""
    import pandas as pd
    from dask import dataframe as dd

    return dd.from_pandas(
        pd.DataFrame.from_records([[0, 1]], columns=["a", "b"]),
        npartitions=1,
    )


@pytest.fixture
def memory_repository():
    """Setup an in-memory repository and clean it up afterwards."""
    root_dir = "/in-memory-root"
    repository = MemoryRepository(root_dir)

    yield repository
    repository.filesystem.rm(root_dir, recursive=True)


@pytest.fixture
def fake_estimator_cls():
    """A fake estimator that exposes the same API as a sklearn
    estimator so we can test without relying on sklearn.
    """

    class FakeEstimator:
        def __init__(self, params=None):
            if params is None:
                params = {"max_df": 0.75, "lowercase": True, "ngram_range": (1, 2)}

            self.params = params

        def get_params(self):
            return self.params

        def fit(self):
            pass

        def transform(self):
            pass

    return FakeEstimator


@pytest.fixture
def viz_experiments(rubicon_and_project_client):
    """Returns a list of experiments with the parameters, metrics, and dataframes
    required to test the `viz` module.
    """
    rubicon, project = rubicon_and_project_client

    dates = pd.date_range(start="1/1/2010", end="12/1/2020", freq="MS")

    for i in range(0, 10):
        experiment = project.log_experiment(
            commit_hash="1234567",
            model_name="test model name",
            name="test name",
            tags=["test tag"],
        )

        experiment.log_parameter(name="test param 0", value=random.choice([True, False]))
        experiment.log_parameter(name="test param 1", value=random.randrange(2, 10, 2))
        experiment.log_parameter(
            name="test param 2", value=random.choice(["A", "B", "C", "D", "E"])
        )

        experiment.log_metric(name="test metric 0", value=random.random())
        experiment.log_metric(name="test metric 1", value=random.random())

        experiment.log_metric(name="test metric 2", value=[random.random() for _ in range(0, 5)])
        experiment.log_metric(name="test metric 3", value=[random.random() for _ in range(0, 5)])

        data = np.array(
            [
                list(dates),
                np.linspace(random.randint(0, 15000), random.randint(0, 15000), len(dates)),
            ]
        )
        data_df = pd.DataFrame.from_records(data.T, columns=["test x", "test y"])

        experiment.log_dataframe(data_df, name="test dataframe")

    return project.experiments()



import dash_bootstrap_components as dbc
from dash import html


def dropdown_header(dropdown_items, selected_item, left_header_text, right_header_text, id_prefix):
    dropdown_menu = dbc.DropdownMenu(
        [
            dbc.DropdownMenuItem(
                item,
                id={
                    "type": f"{id_prefix}-dropdown-button",
                    "index": item,
                },
            )
            for item in dropdown_items
        ],
        color="secondary",
        id=f"{id_prefix}-dropdown",
        label=selected_item,
        size="lg",
    )

    return dbc.Row(
        [
            dbc.Col(
                html.H5(left_header_text, className="header-text"),
                className="header-left-col",
                id=f"{id_prefix}-header-left-col",
                width="auto",
            ),
            dbc.Col(
                dropdown_menu,
                className="header-dropdown-col",
                id=f"{id_prefix}-header-dropdown-col",
                width="auto",
            ),
            dbc.Col(
                html.H5(
                    right_header_text,
                    className="header-text",
                    id=f"{id_prefix}-header-right-text",
                ),
                className="header-right-col",
                id=f"{id_prefix}-header-right-col",
            ),
        ],
        className="header-row",
        id=f"{id_prefix}-header-row",
    )



import pickle

import fsspec

from rubicon_ml.repository import LocalRepository


class MemoryRepository(LocalRepository):
    """The memory repository persists Rubicon data to an
    in-memory filesystem representation.

    The memory repository is intended for testing and
    development purposes. Data persisted to the memory
    filesystem will not be persisted between runtimes.

    Parameters
    ----------
    root_dir : str, optional
        A path representing the virtual root directory for
        the in-memory filesystem. This does not need to be
        specified unless interacting with an already created
        in-memory filesystem.
    storage_options : dict, optional
        Additional keyword arguments that are passed directly to
        the underlying filesystem class.
    """

    PROTOCOL = "memory"

    def __init__(self, root_dir=None, **storage_options):
        self.filesystem = fsspec.filesystem(self.PROTOCOL, **storage_options)
        self.root_dir = root_dir.rstrip("/") if root_dir is not None else "/root"

        if not self.filesystem.exists(self.root_dir):
            self.filesystem.mkdir(self.root_dir)

    def _persist_dataframe(self, df, path):
        """Persists the `dask` dataframe `df` to the in-memory
        path defined by `path`.
        """
        with self.filesystem.open(path, "wb") as f:
            pickle.dump(df, f)

    def _read_dataframe(self, path, df_type="pandas"):
        """Reads the dataframe from the in-memory
        path defined by `path`.
        """
        with self.filesystem.open(path, "rb") as f:
            data = pickle.load(f)

        return data



import fsspec
import yaml


def publish(
    experiments,
    output_filepath=None,
    base_catalog_filepath=None,
):
    """Publish experiments to an `intake` catalog that can be
    read by the `intake-rubicon` driver.

    Parameters
    ----------
    experiments : list of rubicon_ml.client.experiment.Experiment
        The experiments to publish.
    output_filepath : str, optional
        The absolute or relative local filepath or S3 bucket
        and key to log the generated YAML file to. S3 buckets
        must be prepended with 's3://'. Defaults to None,
        which disables writing the generated YAML.
    base_catalog_filepath : str, optional
        Similar to `output_filepath` except this argument is used as a
        base base file to update an existing intake catalog. Defaults to None,
        creating a new intake catalog.

    Returns
    -------
    str
        The YAML string representation of the `intake` catalog
        containing the experiments `experiments`.
    """

    if base_catalog_filepath is not None:
        return _update_catalog(
            base_catalog_filepath=base_catalog_filepath,
            new_experiments=experiments,
            output_filepath=output_filepath,
        )

    catalog = _build_catalog(experiments=experiments)
    catalog_yaml = yaml.dump(catalog)

    if output_filepath is not None:
        with fsspec.open(output_filepath, "w", auto_mkdir=False) as f:
            f.write(catalog_yaml)

    return catalog_yaml


def _update_catalog(base_catalog_filepath, new_experiments, output_filepath=None):

    """Helper function to update exisiting intake catalog.

    Parameters
    ----------
    base_catalog_filepath : str
        the absolute or relative catalog filepath or S3 bucket
        and key to log the generated YAML file to. S3 buckets
        must be prepended with 's3://. Retrieved from the parameter
        of the publish function. NOT optional
    new_experiments : list of rubicon_ml.client.experiment.Experiment
         The new experiments to append to the catalog at
        `base_catalog_filepath`.
    output_catalog_filepath : str, optional
        absolute or relative filepath or S3 bucket
        and key to log the generated YAML file to. (S3 buckets
        must be prepended with 's3://) to  output the updated catalog into.
        Default is None, which resolves to dumping updated catalog into
        base_catalog_filepath path (primary use-case)

    Returns
    -------
    dict
        The dictionary of all sources given as experiments to eventually publish
    """

    updated_catalog = _build_catalog(experiments=new_experiments)

    with fsspec.open(base_catalog_filepath, "r") as yamlfile:
        curr_catalog = yaml.safe_load(yamlfile)

        curr_catalog["sources"].update(updated_catalog["sources"])

    resulting_filepath = base_catalog_filepath if not output_filepath else output_filepath

    with fsspec.open(resulting_filepath, "w") as yamlfile:
        yaml.safe_dump(curr_catalog, yamlfile)
        updated_catalog = yaml.dump(curr_catalog)

    return updated_catalog


def _build_catalog(experiments):

    """Helper function to build catalog dictionary from given experiments.

    Parameters
    ----------
    experiments : list of rubicon_ml.client.experiment.Experiment
        The expriments that are used to build the catalog to eventually publish
    Returns
    -------
    str
        The YAML string representation of the `intake` catalog
        containing the experiments `experiments`.
    """

    catalog = {"sources": {}}

    for experiment in experiments:
        appended_experiment_catalog = {
            "driver": "rubicon_ml_experiment",
            "args": {
                "experiment_id": experiment.id,
                "project_name": experiment.project.name,
                "urlpath": experiment.repository.root_dir,
            },
        }

        experiment_catalog_name = f"experiment_{experiment.id.replace('-', '_')}"
        catalog["sources"][experiment_catalog_name] = appended_experiment_catalog

    return catalog



import warnings


def log_parameter_with_warning(experiment, name, value):
    try:
        experiment.log_parameter(name=name, value=value)
    except Exception as e:
        warning = (
            f"{str(e)}: "
            f"failed to write parameter '{name}' with value {value} of type {type(value)}. "
            f"try using the `FilterEstimatorLogger` with `ignore=['{name}']`"
        )

        warnings.warn(warning)



import os
import pickle
import warnings

import fsspec

from rubicon_ml.client.base import Base
from rubicon_ml.client.mixin import TagMixin
from rubicon_ml.client.utils.exception_handling import failsafe


class Artifact(Base, TagMixin):
    """A client artifact.

    An `artifact` is a catch-all for any other type of
    data that can be logged to a file.

    For example, a snapshot of a trained model (.pkl) can
    be logged to the `experiment` created during its run.
    Or, a base model for the model in development can be
    logged to a `project` when leveraging transfer learning.

    An `artifact` is logged to a `project` or an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Artifact
        The artifact domain model.
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment that the artifact is
        logged to.
    """

    def __init__(self, domain, parent):
        super().__init__(domain, parent._config)

        self._data = None
        self._parent = parent

    def _get_data(self):
        """Loads the data associated with this artifact."""
        project_name, experiment_id = self.parent._get_identifiers()

        self._data = self.repository.get_artifact_data(
            project_name, self.id, experiment_id=experiment_id
        )

    @failsafe
    def get_data(self, unpickle=False):
        """Loads the data associated with this artifact and
        unpickles if needed.

        Parameters
        ----------
        unpickle : bool, optional
            Flag indicating whether artifact data must be
            unpickled. Will be returned as bytes by default.
        """
        project_name, experiment_id = self.parent._get_identifiers()

        data = self.repository.get_artifact_data(project_name, self.id, experiment_id=experiment_id)
        if unpickle:
            data = pickle.loads(data)

        return data

    @failsafe
    def download(self, location=None, name=None):
        """Download this artifact's data.

        Parameters
        ----------
        location : str, optional
            The absolute or relative local directory or S3
            bucket to download the artifact to. S3 buckets
            must be prepended with 's3://'. Defaults to the
            current local working directory.
        name : str, optional
            The name to give the downloaded artifact file.
            Defaults to the artifact's given name when logged.
        """
        if location is None:
            location = os.getcwd()

        if name is None:
            name = self._domain.name

        with fsspec.open(os.path.join(location, name), "wb", auto_mkdir=False) as f:
            f.write(self.data)

    @property
    def id(self):
        """Get the artifact's id."""
        return self._domain.id

    @property
    def name(self):
        """Get the artifact's name."""
        return self._domain.name

    @property
    def description(self):
        """Get the artifact's description."""
        return self._domain.description

    @property
    def created_at(self):
        """Get the time this dataframe was created."""
        return self._domain.created_at

    @property
    def data(self):
        """Get the artifact's raw data."""
        warnings.warn(
            "`data` is deprecated, use `get_data()` instead",
            DeprecationWarning,
        )
        if self._data is None:
            self._get_data()

        return self._data

    @property
    def parent(self):
        """Get the artifact's parent client object."""
        return self._parent



class Base:
    """The base object for all top-level client objects.

    Parameters
    ----------
    domain : one of rubicon.domain.*
        The top-level object's domain instance.
    config : rubicon.client.Config, optional
        The config, which injects the repository to use.
    """

    def __init__(self, domain, config=None):
        self._config = config
        self._domain = domain

    def __str__(self):
        return self._domain.__str__()

    @property
    def repository(self):
        return self._config.repository



import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
from dash import dcc, html
from dash.dependencies import Input, Output

from rubicon_ml.viz.base import VizBase
from rubicon_ml.viz.common.colors import (
    get_rubicon_colorscale,
    light_blue,
    plot_background_blue,
)


class DataframePlot(VizBase):
    """Plot the dataframes with name `dataframe_name` logged to the
    experiments `experiments` on a shared axis.

    Parameters
    ----------
    dataframe_name : str
        The name of the dataframe to plot. A dataframe with name
        `dataframe_name` must be logged to each experiment in `experiments`.
    experiments : list of rubicon_ml.client.experiment.Experiment, optional
        The experiments to visualize. Defaults to None. Can be set as
        attribute after instantiation.
    plotting_func : function, optional
        The `plotly.express` plotting function used to visualize the
        dataframes. Available options can be found at
        https://plotly.com/python-api-reference/plotly.express.html.
        Defaults to `plotly.express.line`.
    plotting_func_kwargs : dict, optional
        Keyword arguments to be passed to `plotting_func`. Available options
        can be found in the documentation of the individual functions at the
        URL above.
    x : str, optional
        The name of the column in the dataframes with name `dataframe_name`
        to plot across the x-axis.
    y : str, optional
        The name of the column in the dataframes with name `dataframe_name`
        to plot across the y-axis.
    """

    def __init__(
        self,
        dataframe_name,
        experiments=None,
        plotting_func=px.line,
        plotting_func_kwargs={},
        x=None,
        y=None,
    ):
        super().__init__(dash_title="plot dataframes")

        self.dataframe_name = dataframe_name
        self.experiments = experiments
        self.plotting_func = plotting_func
        self.plotting_func_kwargs = plotting_func_kwargs
        self.x = x
        self.y = y

    @property
    def layout(self):
        """Defines the dataframe plot's layout."""
        header_text = (
            f"showing dataframe '{self.dataframe_name}' "
            f"over {len(self.experiments)} experiment"
            f"{'s' if len(self.experiments) != 1 else ''}"
        )

        return html.Div(
            [
                html.Div(id="dummy-callback-trigger"),
                dbc.Row(
                    html.H5(header_text, id="header-text"),
                    className="header-row",
                ),
                dcc.Loading(dcc.Graph(id="dataframe-plot"), color=light_blue),
            ],
            id="dataframe-plot-layout-container",
        )

    def load_experiment_data(self):
        """Load the experiment data required for the dataframe plot.

        Extracts the dataframe with name `self.dataframe_name` from
        each experiment in `self.experiment` and combines the data
        stored in them into one dataframe. All dataframes with name
        `dataframe_name` must have the same schema.
        """
        self.data_df = None

        for experiment in self.experiments:
            dataframe = experiment.dataframe(name=self.dataframe_name)

            data_df = dataframe.get_data()
            data_df["experiment_id"] = experiment.id

            if self.x is None:
                self.x = data_df.columns[0]

            if self.y is None:
                self.y = data_df.columns[1]

            if self.data_df is None:
                self.data_df = data_df
            else:
                self.data_df = pd.concat([self.data_df, data_df])

            self.data_df = self.data_df.reset_index(drop=True)

        if "color" not in self.plotting_func_kwargs:
            self.plotting_func_kwargs["color"] = "experiment_id"
        if "color_discrete_sequence" not in self.plotting_func_kwargs:
            self.plotting_func_kwargs["color_discrete_sequence"] = get_rubicon_colorscale(
                len(self.experiments),
            )

    def register_callbacks(self, link_experiment_table=False):
        outputs = [
            Output("dataframe-plot", "figure"),
            Output("header-text", "children"),
        ]
        inputs = [Input("dummy-callback-trigger", "children")]
        states = []

        if link_experiment_table:
            inputs.append(
                Input("experiment-table", "derived_virtual_selected_row_ids"),
            )

        @self.app.callback(outputs, inputs, states)
        def update_dataframe_plot(*args):
            """Render the plot specified by `self.plotting_func`.

            Returns the Plotly figure generated by calling `self.plotting_func`
            on the data in the experiments' dataframes and the header text
            with the dataframes' name.
            """
            if link_experiment_table:
                selected_row_ids = args[-1]
                selected_row_ids = selected_row_ids if selected_row_ids else []
            else:
                selected_row_ids = [e.id for e in self.experiments]

            df_figure_margin = 30

            df_figure = self.plotting_func(
                self.data_df[self.data_df["experiment_id"].isin(selected_row_ids)],
                self.x,
                self.y,
                **self.plotting_func_kwargs,
            )
            df_figure.update_layout(margin_t=df_figure_margin, plot_bgcolor=plot_background_blue)

            for i in range(len(df_figure.data)):
                df_figure.data[i].name = df_figure.data[i].name[:7]

            header_text = (
                f"showing dataframe '{self.dataframe_name}' "
                f"over {len(selected_row_ids)} experiment"
                f"{'s' if len(selected_row_ids) != 1 else ''}"
            )

            return df_figure, header_text



import os
import threading
import time

import dash_bootstrap_components as dbc
from dash import Dash, html

from rubicon_ml import __version__ as rubicon_ml_version

_next_available_port = 8050


class VizBase:
    """The base class for all `rubicon_ml` visualizations.

    `VizBase` can not be directly instantatied. New widgets must all
    extend `VizBase`.
    """

    def __init__(
        self,
        dash_title="base",
    ):
        self.dash_title = f"rubicon-ml: {dash_title}"

    @property
    def layout(self):
        raise NotImplementedError("extensions of `VizBase` must implement property `layout(self)`")

    def build_layout(self):
        """Wraps the layout defined by `self.layout` in a container providing
        the `rubicon_ml` header.
        """
        self.app.layout = dbc.Card(
            dbc.CardBody(
                [
                    dbc.Row(
                        [
                            html.Img(
                                id="rubicon-logo-img",
                                src=self.app.get_asset_url("images/rubicon-logo-dark.png"),
                            ),
                        ],
                    ),
                    dbc.Row(html.P(rubicon_ml_version, id="version-text"), id="version-row"),
                    dbc.Row(self.layout),
                ],
                id="frame",
            ),
        )

    def load_experiment_data(self):
        raise NotImplementedError(
            "extensions of `VizBase` must implement `load_experiment_data(self)`"
        )

    def register_callbacks(self, link_experiment_table=False):
        raise NotImplementedError(
            "extensions of `VizBase` must implement `register_callbacks(self)`"
        )

    def serve(self, in_background=False, dash_kwargs={}, run_server_kwargs={}):
        """Serve the Dash app on the next available port to render the visualization.

        Parameters
        ----------
        in_background : bool, optional
            True to run the Dash app on a thread and return execution to the
            interpreter. False to run the Dash app inline and block execution.
            Defaults to False.
        dash_kwargs : dict, optional
            Keyword arguments to be passed along to the newly instantiated
            Dash object. Available options can be found at
            https://dash.plotly.com/reference#dash.dash.
        run_server_kwargs : dict, optional
            Keyword arguments to be passed along to `Dash.run_server`.
            Available options can be found at
            https://dash.plotly.com/reference#app.run_server. Most commonly,
            the 'port' argument can be provided here to serve the app on a
            specific port.
        """
        if self.experiments is None:
            raise RuntimeError(
                f"`{self.__class__}.experiments` can not be None when `serve` is called"
            )

        self.app = Dash(
            __name__,
            external_stylesheets=[dbc.themes.LUX, dbc.icons.BOOTSTRAP],
            title=self.dash_title,
            **dash_kwargs,
        )

        self.load_experiment_data()
        self.build_layout()
        self.register_callbacks()

        global _next_available_port

        default_run_server_kwargs = {
            "dev_tools_silence_routes_logging": True,
            "port": _next_available_port,
        }
        default_run_server_kwargs.update(run_server_kwargs)

        _next_available_port = default_run_server_kwargs["port"] + 1

        if in_background:
            running_server_thread = threading.Thread(
                name="run_server",
                target=self.app.run_server,
                kwargs=default_run_server_kwargs,
            )
            running_server_thread.daemon = True
            running_server_thread.start()

            port = default_run_server_kwargs.get("port")
            if "proxy" in run_server_kwargs:
                host = default_run_server_kwargs.get("proxy").split("::")[-1]
            else:
                host = f"http://localhost:{port}"

            time.sleep(0.1)  # wait for thread to see if requested port is available
            if not running_server_thread.is_alive():
                raise RuntimeError(f"port {port} may already be in use")

            return host
        else:
            self.app.run_server(**default_run_server_kwargs)

    def show(self, i_frame_kwargs={}, dash_kwargs={}, run_server_kwargs={}):
        """Show the Dash app inline in a Jupyter notebook.

        Parameters
        ----------
        i_frame_kwargs : dict, optional
            Keyword arguments to be passed along to the newly instantiated
            IFrame object. Available options include 'height' and 'width'.
        dash_kwargs : dict, optional
            Keyword arguments to be passed along to the newly instantiated
            Dash object. Available options can be found at
            https://dash.plotly.com/reference#dash.dash.
        run_server_kwargs : dict, optional
            Keyword arguments to be passed along to `Dash.run_server`.
            Available options can be found at
            https://dash.plotly.com/reference#app.run_server. Most commonly,
            the 'port' argument can be provided here to serve the app on a
            specific port.
        """
        from IPython.display import IFrame

        host = self.serve(
            in_background=True, dash_kwargs=dash_kwargs, run_server_kwargs=run_server_kwargs
        )
        proxied_host = os.path.join(host, self.app.config["requests_pathname_prefix"].lstrip("/"))

        default_i_frame_kwargs = {
            "height": "600px",
            "width": "100%",
        }
        default_i_frame_kwargs.update(i_frame_kwargs)

        return IFrame(proxied_host, **default_i_frame_kwargs)



class TagMixin:
    """Adds tagging support to a domain model."""

    def add_tags(self, tags):
        self.tags = list(set(self.tags).union(set(tags)))

    def remove_tags(self, tags):
        self.tags = list(set(self.tags).difference(set(tags)))



from rubicon_ml.repository.base import BaseRepository
from rubicon_ml.repository.local import LocalRepository
from rubicon_ml.repository.memory import MemoryRepository
from rubicon_ml.repository.s3 import S3Repository

__all__ = [
    "BaseRepository",
    "LocalRepository",
    "MemoryRepository",
    "S3Repository",
]



from rubicon_ml.sklearn.filter_estimator_logger import FilterEstimatorLogger
from rubicon_ml.sklearn.pipeline import RubiconPipeline, make_pipeline

__all__ = ["FilterEstimatorLogger", "RubiconPipeline", "make_pipeline"]



from intake.source import base

from rubicon_ml import __version__


class DataSourceMixin(base.DataSource):
    """The base for all `rubicon` Intake data sources.

    Parameters
    ----------
    urlpath : str
        The root directory the `rubicon` data is logged to.
    project_name : str
        The name of the `rubicon` project to load.
    """

    version = __version__

    container = "python"
    name = "rubicon_ml"

    def __init__(self, urlpath, project_name, metadata=None, storage_options=None, **kwargs):
        self._urlpath = urlpath
        self._project_name = project_name
        self._metadata = metadata or {}
        self._storage_options = storage_options or {}
        self._kwargs = kwargs or {}

        super().__init__(metadata=metadata)

    def _get_schema(self):
        """Load the specified `rubicon` object."""
        self._schema = base.Schema(
            datashape=None,
            dtype=None,
            shape=None,
            npartitions=None,
            extra_metadata=self._metadata,
        )

        return self._schema

    def read(self):
        return self._rubicon_object

    def _close(self):
        self._rubicon = None
        self._rubicon_object = None



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import List

from rubicon_ml.domain.mixin import TagMixin
from rubicon_ml.domain.utils import uuid


@dataclass
class Dataframe(TagMixin):
    id: str = field(default_factory=uuid.uuid4)
    name: str = None
    description: str = None
    tags: List[str] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.utcnow)

    parent_id: str = None



def has_tag_requirements(tags, required_tags, qtype):
    """Returns True if `tags` meets the requirements based on
    the values of `required_tags` and `qtype`. False otherwise.
    """
    has_tag_requirements = False

    tag_intersection = set(required_tags).intersection(set(tags))
    if qtype == "or":
        if len(tag_intersection) > 0:
            has_tag_requirements = True
    if qtype == "and":
        if len(tag_intersection) == len(required_tags):
            has_tag_requirements = True

    return has_tag_requirements


def filter_children(children, tags, qtype, name):
    """Filters the provided rubicon objects by `tags` using
    query type `qtype` and by `name`.
    """
    filtered_children = children

    if len(tags) > 0:
        filtered_children = [
            c for c in filtered_children if has_tag_requirements(c.tags, tags, qtype)
        ]
    if name is not None:
        filtered_children = [c for c in filtered_children if c.name == name]

    return filtered_children



from __future__ import annotations

from rubicon_ml.domain.artifact import Artifact
from rubicon_ml.domain.dataframe import Dataframe
from rubicon_ml.domain.experiment import Experiment
from rubicon_ml.domain.feature import Feature
from rubicon_ml.domain.metric import Metric
from rubicon_ml.domain.parameter import Parameter
from rubicon_ml.domain.project import Project

__all__ = ["Artifact", "Dataframe", "Experiment", "Feature", "Metric", "Parameter", "Project"]



from dataclasses import dataclass, field
from datetime import datetime
from typing import List

from rubicon_ml.domain.mixin import TagMixin
from rubicon_ml.domain.utils import uuid


@dataclass
class Parameter(TagMixin):
    name: str

    id: str = field(default_factory=uuid.uuid4)
    value: object = None
    description: str = None
    tags: List[str] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.utcnow)



"""Setup file for the package. For configuration information, see the ``setup.cfg``."""

from setuptools import setup

import versioneer

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



import os
import pickle
import subprocess
import warnings
from datetime import datetime

import fsspec

from rubicon_ml import client, domain
from rubicon_ml.client.utils.exception_handling import failsafe
from rubicon_ml.client.utils.tags import filter_children
from rubicon_ml.exceptions import RubiconException


class ArtifactMixin:
    """Adds artifact support to a client object."""

    def _validate_data(self, data_bytes, data_file, data_object, data_path, name):
        """Raises a `RubiconException` if the data to log as
        an artifact is improperly provided.
        """
        if not any([data_bytes, data_file, data_object, data_path]):
            raise RubiconException(
                "One of `data_bytes`, `data_file`, `data_object` or `data_path` must be provided."
            )

        if name is None:
            if data_path is not None:
                name = os.path.basename(data_path)
            else:
                raise RubiconException("`name` must be provided if not using `data_path`.")

        if data_bytes is None:
            if data_object is not None:
                data_bytes = pickle.dumps(data_object)
            else:
                if data_file is not None:
                    f = data_file
                elif data_path is not None:
                    f = fsspec.open(data_path, "rb")

                with f as open_file:
                    data_bytes = open_file.read()

        return data_bytes, name

    @failsafe
    def log_artifact(
        self,
        data_bytes=None,
        data_file=None,
        data_object=None,
        data_path=None,
        name=None,
        description=None,
        tags=[],
    ):
        """Log an artifact to this client object.

        Parameters
        ----------
        data_bytes : bytes, optional
            The raw bytes to log as an artifact.
        data_file : TextIOWrapper, optional
            The open file to log as an artifact.
        data_object : python object, optional
            The python object to log as an artifact.
        data_path : str, optional
            The absolute or relative local path or S3 path
            to the data to log as an artifact. S3 paths
            must be prepended with 's3://'.
        name : str, optional
            The name of the artifact file. Required if
            `data_path` is not provided.
        description : str, optional
            A description of the artifact. Use to provide
            additional context.
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your artifacts.

        Notes
        -----
        Only one of `data_bytes`, `data_file`, `data_object`, and `data_path`
        should be provided. If more than one is given, the order
        of precedence is `data_bytes`, `data_object`, `data_file`, `data_path`.

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.

        Examples
        --------
        >>> # Log with bytes
        >>> experiment.log_artifact(
        ...     data_bytes=b'hello rubicon!', name='bytes_artifact', description="log artifact from bytes"
        ... )

        >>> # Log with file
        >>> with open('some_relevant_file', 'rb') as f:
        >>>     project.log_artifact(
        ...         data_file=f, name='file_artifact', description="log artifact from file"
        ... )

        >>> # Log with file path
        >>> experiment.log_artifact(
        ...     data_path="./path/to/artifact.pkl", description="log artifact from file path"
        ... )
        """

        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        data_bytes, name = self._validate_data(data_bytes, data_file, data_object, data_path, name)

        artifact = domain.Artifact(
            name=name,
            description=description,
            parent_id=self._domain.id,
            tags=tags,
        )

        project_name, experiment_id = self._get_identifiers()
        self.repository.create_artifact(
            artifact, data_bytes, project_name, experiment_id=experiment_id
        )

        return client.Artifact(artifact, self)

    def _get_environment_bytes(self, export_cmd):
        """Get the working environment as a sequence of bytes.

        Parameters
        ----------
        export_cmd : list of str
            The command to export the environment.

        Returns
        -------
        bytes
            A bytes sequence of the environment.
        """
        try:
            completed_process = subprocess.run(export_cmd, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            raise RubiconException(e.stderr)

        return completed_process.stdout

    @failsafe
    def log_conda_environment(self, artifact_name=None):
        """Log the conda environment as an artifact to this client object.
        Useful for recreating your exact environment at a later date.

        Parameters
        ----------
        artifact_name : str, optional
            The name of the artifact (the exported conda environment).

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.

        Notes
        -----
        Relies on running with an active conda environment.
        """
        if artifact_name is None:
            artifact_name = f"environment-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.yml"

        env_bytes = self._get_environment_bytes(["conda", "env", "export"])
        artifact = self.log_artifact(data_bytes=env_bytes, name=artifact_name)

        return artifact

    @failsafe
    def log_pip_requirements(self, artifact_name=None):
        """Log the pip requirements as an artifact to this client object.
        Useful for recreating your exact environment at a later date.

        Parameters
        ----------
        artifact_name : str, optional
            The name of the artifact (the exported pip environment).

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.
        """
        if artifact_name is None:
            artifact_name = f"requirements-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.txt"

        requirements_bytes = self._get_environment_bytes(["pip", "freeze"])
        artifact = self.log_artifact(data_bytes=requirements_bytes, name=artifact_name)

        return artifact

    @failsafe
    def artifacts(self, name=None, tags=[], qtype="or"):
        """Get the artifacts logged to this client object.

        Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Artifact
            The artifacts previously logged to this client object.
        """
        project_name, experiment_id = self._get_identifiers()
        artifacts = [
            client.Artifact(a, self)
            for a in self.repository.get_artifacts_metadata(
                project_name, experiment_id=experiment_id
            )
        ]

        self._artifacts = filter_children(artifacts, tags, qtype, name)

        return self._artifacts

    @failsafe
    def artifact(self, name=None, id=None):
        """Get an artifact logged to this project by id or name.

        Parameters
        ----------
        id : str
            The id of the artifact to get.
        name : str
            The name of the artifact to get.

        Returns
        -------
        rubicon.client.Artifact
            The artifact logged to this project with id `id` or name 'name'.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            artifacts = self.artifacts(name=name)

            if len(artifacts) == 0:
                raise RubiconException(f"No artifact found with name '{name}'.")
            if len(artifacts) > 1:
                warnings.warn(
                    f"Multiple artifacts found with name '{name}'. Returning most recently logged."
                )

            artifact = artifacts[-1]
        else:
            project_name, experiment_id = self._get_identifiers()
            artifact = client.Artifact(
                self.repository.get_artifact_metadata(project_name, id, experiment_id), self
            )

        return artifact

    @failsafe
    def delete_artifacts(self, ids):
        """Delete the artifacts logged to with client object
        with ids `ids`.

        Parameters
        ----------
        ids : list of str
            The ids of the artifacts to delete.
        """
        project_name, experiment_id = self._get_identifiers()

        for artifact_id in ids:
            self.repository.delete_artifact(project_name, artifact_id, experiment_id=experiment_id)


class DataframeMixin:
    """Adds dataframe support to a client object."""

    @failsafe
    def log_dataframe(self, df, description=None, name=None, tags=[]):
        """Log a dataframe to this client object.

        Parameters
        ----------
        df : pandas.DataFrame or dask.dataframe.DataFrame
            The `dask` or `pandas` dataframe to log.
        description : str, optional
            The dataframe's description. Use to provide
            additional context.
        tags : list of str
            The values to tag the dataframe with.

        Returns
        -------
        rubicon.client.Dataframe
            The new dataframe.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        dataframe = domain.Dataframe(
            parent_id=self._domain.id,
            description=description,
            name=name,
            tags=tags,
        )

        project_name, experiment_id = self._get_identifiers()
        self.repository.create_dataframe(dataframe, df, project_name, experiment_id=experiment_id)

        return client.Dataframe(dataframe, self)

    @failsafe
    def dataframes(self, name=None, tags=[], qtype="or"):
        """Get the dataframes logged to this client object.

        Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Dataframe
            The dataframes previously logged to this client object.
        """
        project_name, experiment_id = self._get_identifiers()
        dataframes = [
            client.Dataframe(d, self)
            for d in self.repository.get_dataframes_metadata(
                project_name, experiment_id=experiment_id
            )
        ]

        self._dataframes = filter_children(dataframes, tags, qtype, name)

        return self._dataframes

    @failsafe
    def dataframe(self, name=None, id=None):
        """
        Get the dataframe logged to this client object.

        Parameters
        ----------
        id : str
            The id of the dataframe to get.
        name : str
            The name of the dataframe to get.
        Returns
        -------
        rubicon.client.Dataframe
            The dataframe logged to this project with id `id` or name 'name'.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        elif name is not None:
            dataframes = self.dataframes(name=name)

            if len(dataframes) == 0:
                raise RubiconException(f"No dataframe found with name '{name}'.")
            elif len(dataframes) > 1:
                warnings.warn(
                    f"Multiple dataframes found with name '{name}'."
                    " Returning most recently logged."
                )

            dataframe = dataframes[-1]
        else:
            project_name, experiment_id = self._get_identifiers()
            dataframe = client.Dataframe(
                self.repository.get_dataframe_metadata(
                    project_name, experiment_id=experiment_id, dataframe_id=id
                ),
                self,
            )

        return dataframe

    @failsafe
    def delete_dataframes(self, ids):
        """Delete the dataframes with ids `ids` logged to
        this client object.

        Parameters
        ----------
        ids : list of str
            The ids of the dataframes to delete.
        """
        project_name, experiment_id = self._get_identifiers()

        for dataframe_id in ids:
            self.repository.delete_dataframe(
                project_name, dataframe_id, experiment_id=experiment_id
            )


class TagMixin:
    """Adds tag support to a client object."""

    def _get_taggable_identifiers(self):
        project_name, experiment_id = self._parent._get_identifiers()
        entity_identifier = None

        # experiments do not return an entity identifier - they are the entity
        if isinstance(self, client.Experiment):
            experiment_id = self.id
        # dataframes and artifacts are identified by their `id`s
        elif isinstance(self, client.Dataframe) or isinstance(self, client.Artifact):
            entity_identifier = self.id
        # everything else is identified by its `name`
        else:
            entity_identifier = self.name

        return project_name, experiment_id, entity_identifier

    @failsafe
    def add_tags(self, tags):
        """Add tags to this client object.

        Parameters
        ----------
        tags : list of str
            The tag values to add.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        project_name, experiment_id, entity_identifier = self._get_taggable_identifiers()

        self._domain.add_tags(tags)
        self.repository.add_tags(
            project_name,
            tags,
            experiment_id=experiment_id,
            entity_identifier=entity_identifier,
            entity_type=self.__class__.__name__,
        )

    @failsafe
    def remove_tags(self, tags):
        """Remove tags from this client object.

        Parameters
        ----------
        tags : list of str
             The tag values to remove.
        """
        project_name, experiment_id, entity_identifier = self._get_taggable_identifiers()

        self._domain.remove_tags(tags)
        self.repository.remove_tags(
            project_name,
            tags,
            experiment_id=experiment_id,
            entity_identifier=entity_identifier,
            entity_type=self.__class__.__name__,
        )

    def _update_tags(self, tag_data):
        """Add or remove the tags in `tag_data` based on
        their key.
        """
        for tag in tag_data:
            self._domain.add_tags(tag.get("added_tags", []))
            self._domain.remove_tags(tag.get("removed_tags", []))

    @property
    def tags(self):
        """Get this client object's tags."""
        project_name, experiment_id, entity_identifier = self._get_taggable_identifiers()
        tag_data = self.repository.get_tags(
            project_name,
            experiment_id=experiment_id,
            entity_identifier=entity_identifier,
            entity_type=self.__class__.__name__,
        )

        self._update_tags(tag_data)

        return self._domain.tags



import os

import dash_bootstrap_components as dbc
from dash import dash_table, dcc, html
from dash.dependencies import ALL, Input, Output, State

from rubicon_ml import publish
from rubicon_ml.viz.base import VizBase
from rubicon_ml.viz.common.colors import light_blue, plot_background_blue


class ExperimentsTable(VizBase):
    """Visualize the experiments `experiments` and their metadata, metrics,
    and parameters in a tabular format.

    Parameters
    ----------
    experiments : list of rubicon_ml.client.experiment.Experiment, optional
        The experiments to visualize. Defaults to None. Can be set as
        attribute after instantiation.
    is_selectable : bool, optional
        True to enable selection of the rows in the table, False otherwise.
        Defaults to True.
    """

    def __init__(self, experiments=None, is_selectable=True):
        super().__init__(dash_title="experiment table")

        self.experiments = experiments
        self.is_selectable = is_selectable

    @property
    def layout(self):
        """Defines the experiments table's layout."""
        bulk_select_buttons = [
            html.Div(
                dbc.Button(
                    "select all experiments",
                    color="primary",
                    disabled=not self.is_selectable,
                    id="select-all-button",
                    outline=True,
                ),
                className="bulk-select-button-container",
            ),
            html.Div(
                dbc.Button(
                    "clear all experiments",
                    color="primary",
                    disabled=not self.is_selectable,
                    id="clear-all-button",
                    outline=True,
                ),
                className="bulk-select-button-container",
            ),
            html.Div(
                dbc.Button(
                    "publish selected",
                    color="primary",
                    disabled=not self.is_selectable,
                    id="publish-selected-button",
                    outline=True,
                ),
                className="bulk-select-button-container",
            ),
        ]

        experiment_table = dash_table.DataTable(
            columns=[
                {"name": column, "id": column, "selectable": self.is_selectable}
                for column in self.all_columns
            ],
            data=self.experiment_records,
            filter_action="native",
            fixed_columns={"headers": True, "data": 1},
            hidden_columns=self.hidden_columns,
            id="experiment-table",
            page_size=10,
            row_selectable="multi" if self.is_selectable else False,
            selected_rows=[],
            sort_action="native",
            sort_mode="multi",
            style_cell={"overflow": "hidden", "textOverflow": "ellipsis"},
            style_data_conditional=[
                {"if": {"row_index": "odd"}, "backgroundColor": plot_background_blue}
            ],
            style_header={"fontWeight": 700},
            style_table={"minWidth": "100%"},
        )

        header_text = (
            f"showing {len(self.experiments)} experiments "
            f"{'at commit ' if self.commit_hash is not None else ''}"
        )

        header = html.H5(
            [
                html.P(
                    header_text,
                    className="experiment-table-header-text",
                    style={"float": "left"} if self.commit_hash is not None else {},
                ),
                html.A(
                    html.P(
                        [
                            self.commit_hash,
                            html.I(className="bi bi-box-arrow-up-right external-link-icon"),
                        ]
                    ),
                    href=self.github_url,
                    style={"display": "none"} if self.commit_hash is None else {},
                    target="_blank",
                ),
            ],
            className="header-text",
        )

        toggle_columns_dropdown = dbc.DropdownMenu(
            [
                dbc.DropdownMenuItem(
                    "show all",
                    id="show-all-dropdown-button",
                ),
                dbc.DropdownMenuItem(
                    "hide all",
                    id="hide-all-dropdown-button",
                ),
                dbc.DropdownMenuItem(divider=True),
                *[
                    dbc.DropdownMenuItem(
                        dcc.Checklist(
                            inputClassName="column-dropdown-checkbox",
                            labelClassName="column-dropdown-label",
                            options=[{"label": column, "value": column}],
                            value=[column] if column not in self.hidden_columns else [],
                            id={
                                "type": "column-dropdown-checkbox",
                                "index": column,
                            },
                        ),
                        id={"type": "column-dropdown-button", "index": column},
                    )
                    for column in self.all_columns
                ],
            ],
            color="secondary",
            id="column-selection-dropdown",
            label="toggle columns",
        )

        publish_modal = dbc.Modal(
            [
                dbc.ModalHeader(
                    dbc.ModalTitle("publish selected experiments"),
                    close_button=True,
                ),
                dbc.ModalBody(
                    [
                        dbc.Label("enter catalog YAML output path:"),
                        dbc.Input(
                            id="publish-path-input",
                            type="text",
                            value=os.path.join(os.getcwd(), "rubicon-ml-catalog.yml"),
                        ),
                    ],
                ),
                dbc.ModalFooter(dbc.Button("publish", id="publish-button")),
            ],
            id="publish-modal",
            centered=True,
            is_open=False,
            size="lg",
        )

        if self.is_selectable:
            header_row = [
                html.Div(
                    header,
                    className="header-row",
                ),
                dbc.Row(
                    [
                        *[dbc.Col(button, width="auto") for button in bulk_select_buttons],
                        dbc.Col(toggle_columns_dropdown),
                    ],
                    className="button-group",
                ),
            ]
        else:
            header_row = [
                dbc.Row(
                    [
                        dbc.Col(
                            html.Div(
                                header,
                                className="header-row",
                            ),
                            width=8,
                        ),
                        dbc.Col(toggle_columns_dropdown),
                    ],
                    className="button-group",
                ),
            ]

        return html.Div(
            [
                *header_row,
                publish_modal,
                dcc.Loading(experiment_table, color=light_blue),
            ],
        )

    def load_experiment_data(self):
        """Load the experiment data required for the experiments table.

        Extracts all experiment metadata as well as parameters and metrics
        from each experiment in `self.experiments`. Sets GitHub information
        if applicable.
        """
        self.experiment_records = []
        self.metric_names = set()
        self.parameter_names = set()

        self.all_columns = ["id", "name", "created_at", "model_name", "commit_hash", "tags"]
        self.hidden_columns = []

        self.commit_hash = None
        self.github_url = None

        commit_hashes = set()
        show_columns = {"id", "created_at"}

        for experiment in self.experiments:
            experiment_record = {
                "id": experiment.id,
                "name": experiment.name,
                "created_at": experiment.created_at,
                "model_name": experiment.model_name,
                "commit_hash": None,
                "tags": ", ".join(str(tag) for tag in experiment.tags),
            }

            if experiment.commit_hash is not None:
                experiment_record["commit_hash"] = experiment.commit_hash[:7]

                commit_hashes.add(experiment.commit_hash)
                show_columns.add("commit_hash")

            if experiment.model_name is not None:
                show_columns.add("model_name")

            if experiment.name is not None:
                show_columns.add("name")

            if len(experiment.tags) > 0:
                show_columns.add("tags")

            for parameter in experiment.parameters():
                experiment_record[parameter.name] = parameter.value

                self.parameter_names.add(parameter.name)

            for metric in experiment.metrics():
                experiment_record[metric.name] = metric.value

                self.metric_names.add(metric.name)

            self.experiment_records.append(experiment_record)

        self.metric_names = list(self.metric_names)
        self.parameter_names = list(self.parameter_names)

        self.all_columns.extend(self.parameter_names + self.metric_names)
        self.hidden_columns = [
            column
            for column in self.all_columns
            if column not in list(show_columns) + self.metric_names + self.parameter_names
        ]

        if len(commit_hashes) == 1:
            github_url_root = self.experiments[0].project.github_url[:-4]
            commit_hash = list(commit_hashes)[0]

            self.commit_hash = commit_hash[:7]
            self.github_url = f"{github_url_root}/tree/{commit_hash}"

    def register_callbacks(self, link_experiment_table=False):
        @self.app.callback(
            Output({"type": "column-dropdown-checkbox", "index": ALL}, "value"),
            [
                Input("show-all-dropdown-button", "n_clicks_timestamp"),
                Input("hide-all-dropdown-button", "n_clicks_timestamp"),
            ],
            prevent_initial_call=True,
        )
        def update_selected_column_checkboxes(last_show_click, last_hide_click):
            """Bulk updates for the selections in the "toggle columns" dropdown.

            Returns all if triggered by the "show all" button and returns
            only "id" if triggered by the "hide all" button. The initial
            call is prevented as the default state is neither all nor none.
            """
            last_show_click = last_show_click if last_show_click else 0
            last_hide_click = last_hide_click if last_hide_click else 0

            if last_hide_click > last_show_click:
                hidden_values = [[]] * (len(self.all_columns) - 1)

                return [["id"], *hidden_values]

            return [[column] for column in self.all_columns]

        @self.app.callback(
            Output("experiment-table", "hidden_columns"),
            Input({"type": "column-dropdown-checkbox", "index": ALL}, "value"),
        )
        def update_hidden_experiment_table_cols(selected_columns):
            """Hide and show the columns in the experiment table.

            Returns the columns that should be hidden based on whether or
            not the column's corresponding value is checked in the "toggle
            columns" dropdown.
            """
            selected_columns = [sc[0] for sc in selected_columns if len(sc) > 0]

            return [column for column in self.all_columns if column not in selected_columns]

        @self.app.callback(
            Output("experiment-table", "selected_rows"),
            [
                Input("select-all-button", "n_clicks_timestamp"),
                Input("clear-all-button", "n_clicks_timestamp"),
            ],
            State("experiment-table", "derived_virtual_indices"),
        )
        def update_selected_experiment_table_rows(
            last_select_click, last_clear_click, experiment_table_indices
        ):
            """Bulk selection for the rows of the experiment table.

            Returns all if triggered by the "select all" button and returns
            none if triggered by the "clear all" button.
            """
            if last_select_click is None and last_clear_click is None:
                return list(range(len(self.experiments)))

            last_select_click = last_select_click if last_select_click else 0
            last_clear_click = last_clear_click if last_clear_click else 0

            if last_select_click > last_clear_click:
                return experiment_table_indices

            return []

        @self.app.callback(
            Output("publish-modal", "is_open"),
            [
                Input("publish-selected-button", "n_clicks_timestamp"),
                Input("publish-button", "n_clicks_timestamp"),
            ],
            [
                State("publish-modal", "is_open"),
                State("publish-path-input", "value"),
                State("experiment-table", "derived_virtual_selected_rows"),
                State("experiment-table", "derived_virtual_data"),
            ],
        )
        def toggle_publish_modal(
            last_publish_selected_click,
            last_publish_click,
            is_modal_open,
            publish_path,
            selected_rows,
            data,
        ):
            last_publish_selected_click = (
                last_publish_selected_click if last_publish_selected_click else 0
            )
            last_publish_click = last_publish_click if last_publish_click else 0

            if last_publish_selected_click > last_publish_click:
                return True
            elif last_publish_click > last_publish_selected_click:
                selected_experiment_ids = [data[row].get("id") for row in selected_rows]
                selected_experiments = [
                    e for e in self.experiments if e.id in selected_experiment_ids
                ]

                publish(selected_experiments, publish_path)

                return False

            return is_modal_open



import subprocess
import warnings

from rubicon_ml import domain
from rubicon_ml.client import Config, Project
from rubicon_ml.client.utils.exception_handling import failsafe
from rubicon_ml.exceptions import RubiconException
from rubicon_ml.repository.utils import slugify


class Rubicon:
    """The `rubicon` client's entry point.

    Creates a `Config` and injects it into the client level
    objects at run-time.

    Parameters
    ----------
    persistence : str, optional
        The persistence type. Can be one of ["filesystem", "memory"].
        Defaults to "filesystem".
    root_dir : str, optional
        Absolute or relative filepath. Use absolute path for best performance.
        Defaults to the local filesystem. Prefix with s3:// to use s3 instead.
    auto_git_enabled : bool, optional
        True to use the `git` command to automatically log relevant repository
        information to projects and experiments logged with this client instance,
        False otherwise. Defaults to False.
    storage_options : dict, optional
        Additional keyword arguments specific to the protocol being chosen. They
        are passed directly to the underlying filesystem class.
    """

    def __init__(
        self, persistence="filesystem", root_dir=None, auto_git_enabled=False, **storage_options
    ):
        self.config = Config(persistence, root_dir, auto_git_enabled, **storage_options)

    @property
    def repository(self):
        return self.config.repository

    @repository.setter
    def repository(self, value):
        self.config.repository = value

    def _get_github_url(self):
        """Returns the repository URL of the `git` repo it is called from."""
        completed_process = subprocess.run(["git", "remote", "-v"], capture_output=True)
        remotes = completed_process.stdout.decode("utf8").replace("\t", " ").split("\n")

        try:
            origin = [remote for remote in remotes if remote.startswith("origin")][0]
            github_url = origin.split(" ")[1]
        except IndexError:
            github_url = None

        return github_url

    def _create_project_domain(self, name, description, github_url, training_metadata):
        """Instantiates and returns a project domain object."""
        if self.config.is_auto_git_enabled and github_url is None:
            github_url = self._get_github_url()

        if training_metadata is not None:
            training_metadata = domain.utils.TrainingMetadata(training_metadata)

        return domain.Project(
            name,
            description=description,
            github_url=github_url,
            training_metadata=training_metadata,
        )

    @failsafe
    def create_project(self, name, description=None, github_url=None, training_metadata=None):
        """Create a project.

        Parameters
        ----------
        name : str
            The project's name.
        description : str, optional
            The project's description.
        github_url : str, optional
            The URL of the GitHub repository associated with this
            project. If omitted and automatic `git` logging is
            enabled, it will be retrieved via `git remote`.
        training_metadata : tuple or list of tuples, optional
            Metadata associated with the training dataset(s)
            used across each experiment in this project.

        Returns
        -------
        rubicon.client.Project
            The created project.
        """
        project = self._create_project_domain(name, description, github_url, training_metadata)
        self.repository.create_project(project)

        return Project(project, self.config)

    @failsafe
    def get_project(self, name=None, id=None):
        """Get a project.

        Parameters
        ----------
        name : str, optional
            The name of the project to get.
        id : str, optional
            The id of the project to get.

        Returns
        -------
        rubicon.client.Project
            The project with name `name` or id `id`.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            project = self.repository.get_project(name)
            project = Project(project, self.config)
        else:
            project = [p for p in self.projects() if p.id == id][0]

        return project

    def get_project_as_dask_df(self, name, group_by=None):
        """DEPRECATED: Available for backwards compatibility."""
        warnings.warn(
            "`get_project_as_dask_df` is deprecated and will be removed in a future "
            "release. use `get_project_as_df('name', df_type='dask') instead.",
            DeprecationWarning,
        )

        return self.get_project_as_df(name, df_type="dask", group_by=group_by)

    @failsafe
    def get_project_as_df(self, name, df_type="pandas", group_by=None):
        """Get a dask or pandas dataframe representation of a project.

        Parameters
        ----------
        name : str
            The name of the project to get.
        df_type : str, optional
            The type of dataframe to return. Valid options include
            ["dask", "pandas"]. Defaults to "pandas".
        group_by : str or None, optional
            How to group the project's experiments in the returned
            DataFrame(s). Valid options include ["commit_hash"].

        Returns
        -------
        pandas.DataFrame or list of pandas.DataFrame or dask.DataFrame or list of dask.DataFrame
            If `group_by` is `None`, a dask or pandas dataframe holding the project's
            data. Otherwise a list of dask or pandas dataframes holding the project's
            data grouped by `group_by`.
        """
        project = self.get_project(name)

        return project.to_df(df_type=df_type, group_by=None)

    @failsafe
    def get_or_create_project(self, name, **kwargs):
        """Get or create a project.

        Parameters
        ----------
        name : str
            The project's name.
        kwargs : dict
            Additional keyword arguments to be passed to
            `Rubicon.create_project`.

        Returns
        -------
        rubicon.client.Project
            The corresponding project.
        """
        try:
            project = self.get_project(name)
        except RubiconException:
            project = self.create_project(name, **kwargs)
        else:  # check for None in case of failure mode being set to "log" or "warn"
            if project is None:
                project = self.create_project(name, **kwargs)

        return project

    @failsafe
    def projects(self):
        """Get a list of available projects.

        Returns
        -------
        list of rubicon.client.Project
            The list of available projects.
        """
        return [Project(project, self.config) for project in self.repository.get_projects()]

    @failsafe
    def sync(self, project_name, s3_root_dir):
        """Sync a local project to S3.

        Parameters
        ----------
        project_name : str
            The name of the project to sync.
        s3_root_dir : str
            The S3 path where the project's data
            will be synced.

        Notes
        -----
        Use to backup your local project data to S3, as an alternative to direct S3 logging.
        Relies on AWS CLI's sync. Ensure that your credentials are set and that your Proxy
        is on.
        """
        if self.config.persistence != "filesystem":
            raise RubiconException(
                "You can't sync projects written to memory. Sync from either local filesystem or S3."
            )

        project = self.get_project(project_name)
        local_path = f"{self.config.root_dir}/{slugify(project.name)}"
        cmd = f"aws s3 sync {local_path} {s3_root_dir}/{slugify(project.name)}"

        try:
            subprocess.run(cmd, shell=True, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            raise RubiconException(e.stderr)



from rubicon_ml import domain
from rubicon_ml.client import (
    ArtifactMixin,
    Base,
    DataframeMixin,
    Feature,
    Metric,
    Parameter,
    TagMixin,
)
from rubicon_ml.client.utils.exception_handling import failsafe
from rubicon_ml.client.utils.tags import filter_children


class Experiment(Base, ArtifactMixin, DataframeMixin, TagMixin):
    """A client experiment.

    An `experiment` represents a model run and is identified by
    its 'created_at' time. It can have `metrics`, `parameters`,
    `features`, `dataframes`, and `artifacts` logged to it.

    An `experiment` is logged to a `project`.

    Parameters
    ----------
    domain : rubicon.domain.Experiment
        The experiment domain model.
    parent : rubicon.client.Project
        The project that the experiment is logged to.
    """

    def __init__(self, domain, parent):
        super().__init__(domain, parent._config)

        self._parent = parent
        self._artifacts = []
        self._dataframes = []
        self._metrics = []
        self._features = []
        self._parameters = []

    def _get_identifiers(self):
        """Get the experiment's project's name and the experiment's ID."""
        return self.project.name, self.id

    @failsafe
    def log_metric(self, name, value, directionality="score", description=None, tags=[]):
        """Create a metric under the experiment.

        Parameters
        ----------
        name : str
            The metric's name.
        value : float
            The metric's value.
        directionality : str, optional
            The metric's directionality. Must be one of
            ["score", "loss"], where "score" represents
            a metric to maximize, while "loss" represents a
            metric to minimize. Defaults to "score".
        description : str, optional
            The metric's description. Use to provide additional
            context.
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your metrics.

        Returns
        -------
        rubicon.client.Metric
            The created metric.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        metric = domain.Metric(
            name, value, directionality=directionality, description=description, tags=tags
        )
        self.repository.create_metric(metric, self.project.name, self.id)

        return Metric(metric, self)

    @failsafe
    def metrics(self, name=None, tags=[], qtype="or"):
        """Get the metrics logged to this experiment.

        Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Metric
            The metrics previously logged to this experiment.
        """
        metrics = [Metric(m, self) for m in self.repository.get_metrics(self.project.name, self.id)]
        self._metrics = filter_children(metrics, tags, qtype, name)

        return self._metrics

    @failsafe
    def metric(self, name=None, id=None):
        """Get a metric.

        Parameters
        ----------
        name : str, optional
            The name of the metric to get.
        id : str, optional
            The id of the metric to get.

        Returns
        -------
        rubicon.client.Metric
            The metric with name `name` or id `id`.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            metric = self.repository.get_metric(self.project.name, self.id, name)
            metric = Metric(metric, self)
        else:
            metric = [m for m in self.metrics() if m.id == id][0]

        return metric

    @failsafe
    def log_feature(self, name, description=None, importance=None, tags=[]):
        """Create a feature under the experiment.

        Parameters
        ----------
        name : str
            The features's name.
        description : str
            The feature's description. Use to provide
            additional context.
        importance : float
            The feature's importance.
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your features.

        Returns
        -------
        rubicon.client.Feature
            The created feature.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        feature = domain.Feature(name, description=description, importance=importance, tags=tags)
        self.repository.create_feature(feature, self.project.name, self.id)

        return Feature(feature, self)

    @failsafe
    def features(self, name=None, tags=[], qtype="or"):
        """Get the features logged to this experiment.

         Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Feature
            The features previously logged to this experiment.
        """

        features = [
            Feature(f, self) for f in self.repository.get_features(self.project.name, self.id)
        ]

        self._features = filter_children(features, tags, qtype, name)
        return self._features

    @failsafe
    def feature(self, name=None, id=None):
        """Get a feature.

        Parameters
        ----------
        name : str, optional
            The name of the feature to get.
        id : str, optional
            The id of the feature to get.

        Returns
        -------
        rubicon.client.Feature
            The feature with name `name` or id `id`.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            feature = self.repository.get_feature(self.project.name, self.id, name)
            feature = Feature(feature, self)
        else:
            feature = [f for f in self.features() if f.id == id][0]

        return feature

    @failsafe
    def log_parameter(self, name, value=None, description=None, tags=[]):
        """Create a parameter under the experiment.

        Parameters
        ----------
        name : str
            The parameter's name.
        value : object, optional
            The parameter's value. Can be an object of any JSON
            serializable (via rubicon.utils.DomainJSONEncoder)
            type.
        description : str, optional
            The parameter's description. Use to provide
            additional context.
        tags : list of str, optional
            Values to tag the parameter with. Use tags to organize and
            filter your parameters.

        Returns
        -------
        rubicon.client.Parameter
            The created parameter.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        parameter = domain.Parameter(name, value=value, description=description, tags=tags)
        self.repository.create_parameter(parameter, self.project.name, self.id)

        return Parameter(parameter, self)

    @failsafe
    def parameters(self, name=None, tags=[], qtype="or"):
        """Get the parameters logged to this experiment.

        Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Parameter
            The parameters previously logged to this experiment.
        """

        parameters = [
            Parameter(p, self) for p in self.repository.get_parameters(self.project.name, self.id)
        ]

        self._parameters = filter_children(parameters, tags, qtype, name)

        return self._parameters

    @failsafe
    def parameter(self, name=None, id=None):
        """Get a parameter.

        Parameters
        ----------
        name : str, optional
            The name of the parameter to get.
        id : str, optional
            The id of the parameter to get.

        Returns
        -------
        rubicon.client.Parameter
            The parameter with name `name` or id `id`.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            parameter = self.repository.get_parameter(self.project.name, self.id, name)
            parameter = Parameter(parameter, self)
        else:
            parameter = [p for p in self.parameters() if p.id == id][0]

        return parameter

    @property
    def id(self):
        """Get the experiment's id."""
        return self._domain.id

    @property
    def name(self):
        """Get the experiment's name."""
        return self._domain.name

    @property
    def description(self):
        """Get the experiment's description."""
        return self._domain.description

    @property
    def model_name(self):
        """Get the experiment's model name."""
        return self._domain.model_name

    @property
    def branch_name(self):
        """Get the experiment's branch name."""
        return self._domain.branch_name

    @property
    def commit_hash(self):
        """Get the experiment's commit hash."""
        return self._domain.commit_hash

    @property
    def training_metadata(self):
        """Get the project's training metadata."""
        training_metadata = self._domain.training_metadata.training_metadata

        if len(training_metadata) == 1:
            training_metadata = training_metadata[0]

        return training_metadata

    @property
    def created_at(self):
        """Get the time the experiment was created."""
        return self._domain.created_at

    @property
    def project(self):
        """Get the project client object that this experiment
        belongs to.
        """
        return self._parent



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import List

from rubicon_ml.domain.mixin import TagMixin
from rubicon_ml.domain.utils import TrainingMetadata, uuid


@dataclass
class Experiment(TagMixin):
    project_name: str

    id: str = field(default_factory=uuid.uuid4)
    name: str = None
    description: str = None
    model_name: str = None
    branch_name: str = None
    commit_hash: str = None
    training_metadata: TrainingMetadata = None
    tags: List[str] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.utcnow)
